"Tools for using reflection to inspect function calls."
import inspect


def get_callargs(f, *call_args, **call_kwargs):
    """
    Return a dictionary mapping the names of arguments to f,
    including `*packed_varargs` and `**packed_kwargs` type arguments,
    to their values in the call to f.

    This function is intended to be usable for logging and exception
    handling, so it should never raise, which is part of what
    distinguishes it from the built-in `inspect.get_call_arguments`
    function.

    PARAMETERS
    ----------
    f : function or method
        Note that f cannot necessarily be an arbitrary callable, and
        also that many decorators will clobber the metadata needed. If
        you use `tdxutil.wrapping.wraps` for decorators, the metadata
        should still be available.
    call_args : tuple
        The positional arguments in a call to `f`.
    call_args : dict
        The keyword arguments in a call to `f`.

    RETURNS
    -------
    arguments: dict
        A dict whose keys are string names of arguments in the
        signature of f. All of the regular (non-packed) arguments
        are guaranteed to appear here; if they are missing in the call then
        they will be set to a value such as `"__missing_argument_x__"`. The
        packed varargs argument, if any exists in the signature of `f`, will be
        a possibly empty tuple. The packed kwargs argument, if any exists in
        the signature of `f`, will be a possibly empty dict.

    NOTES
    -----

    The python standard library has a very similar function,
    `inspect.getcallargs`. There are two reasons to use this implementation
    instead:
      * Because this implementation should never raise, it's more suitable
        for use in error-handling code (which is a common use case
        for decorators)
      * In speed tests, `get_call_arguments` seems to be very slightly
        faster, although for reasonably small argument lists both
        functions run in about 5-10 microseconds in tests on a
        2016 macbook pro.

    If there are extra positional or keyword arguments that make the call
    illegal, they will be ingored in the output. If a regular argument is
    specified both positionally and as a keyword argument, the positional
    specification wins in this function. Both of these situations would
    lead to a `TypeError` being raised if the call were actually attempted.

    In order for this method to work, the argspec of the function cannot
    be clobbered. If you are decorating a function with decorators that use
    `functools.wraps` to preserve metadata, the argspec metadata will be lost,
    but if you use `wrapt.decorator` then it will be preserved.

    EXAMPLES
    --------
    >>> def f(a, b, c, d=4, *stuff, **morestuff): pass

    >>> get_call_arguments(f, [1, 2], {'x':3})
    # {'a': 1,
    #  'b': 2,
    #  'c': '__missing_argument_c__',
    #  'd': 4,
    #  'morestuff': {'x': 3},
    #  'stuff': ()}

    >>> get_call_arguments(f, [1, 2, 3, 4, 5, 6], {'x':3})
    # {'a': 1,
    #  'b': 2,
    #  'c': 3,
    #  'd': 4,
    #  'morestuff': {'x': 3},
    #  'stuff': (5, 6)}

    See the tests for more examples.

    """
    f_argspec = inspect.getargspec(f)
    return callargs_from_argspec(f_argspec, *call_args, **call_kwargs)


def callargs_from_argspec(f_argspec, *call_args, **call_kwargs):
    """
    Like get_callargs (see that function for docs) except instead of
    passing in a callable f, we pass in its argspec.

    The reason this function exists is that sometimes we have to do
    an end-run around the usual argspec. For example, in the `clickutil`
    package I have been unable to make the metadata of a click-decorated
    function match that of the original function (wrapt doesn't work
    with click because the decorators are classes, not wrapper unctions,
    so I cannot just wraptify the click decorators). By separating
    out the core logic that depends only on the argspec, I make the
    `tdx` module more flexible and useful in more situations.

    """
    f_args, f_varargs, f_kwargs, f_defaults = get_signature(f_argspec)

    # there are two major cases which are easier to deal with
    # separately:
    #  * if the number of positional args in the call
    #    exceeds the number of regular args of f, then there are
    #    varargs, and also none of f's regular args are specified
    #    via kwargs in the call or default values
    #  * otherwise, there are no varargs, and we have to check
    #    kwargs of the call and default values for some of the
    #    regular arguments of f
    if len(call_args) > len(f_args):
        return _get_arguments_extra_positional(
            call_args, call_kwargs,
            f_args, f_varargs, f_kwargs
        )
    else:
        return _get_arguments_no_extra_positional(
            call_args, call_kwargs,
            f_args, f_defaults, f_varargs, f_kwargs
        )


def get_signature(argspec):
    """
    Return a tuple (args, defaults, varargs, kwargs) from a function
    signature. If there are no varargs, then `varargs` is None.

    PARAMETERS
    ----------
    f : function or method
        Note that f cannot necessarily be an arbitrary callable, and
        also that many decorators will clobber the metadata needed. If
        you use `tdxutil.wrapping.wraps` for decorators, the metadata
        should still be available.

    RETURNS
    -------
    args : list
        A list of the string names of all regular arguments (not unpacked
        varargs or keyword args) of `f`.
    varargs : {None, str}
        If the signature of f contains a varargs positional argument,
        this will be the string name of that argument. (For example if
        the argument list contains `*myvarargs` we get `"myvarargs"``).
        If there is no vararg argument, this will be None.
    kwargs : {None, str}
        If the signature of f contains a packed keyword argument,
        this will be the string name of that argument. (For example if
        the argument list contains `**mykwargs` we get `"mykwargs"``).
        If there is no packed keyword argument, this will be None.
    defaults : list
        A list of all the default values for regular arguments of `f`.
        In general it can be shorter than `args`, and `defaults` will
        be aligned with `args[-len(defaults):]`

    EXAMPLES
    --------
    >>> def f(a, b=3, c=5, *stuff, **morestuff): pass

    >>> get_signature(f)
    # (['a', 'b', 'c'], [3, 5], 'stuff', 'morestuff')

    >>> def f(x, y): pass

    >>> get_signature(f)
    # (['x', 'y'], [], None, None)
    """
    f_args = argspec.args
    f_varargs = argspec.varargs
    f_kwargs = argspec.keywords
    f_defaults = [] if argspec.defaults is None else list(argspec.defaults)
    return f_args, f_varargs, f_kwargs, f_defaults


def _get_arguments_extra_positional(call_args, call_kwargs,
                                    f_args, f_varargs, f_kwargs,):
    arguments = {}
    # handle all of the regular args of f
    n_f_args = len(f_args)
    for name, value in zip(f_args, call_args):
        arguments[name] = value
    # set the f_varargs key
    # -------------------
    # we need the check on f_varargs because if the user made an argument
    # mismatch error, then f may not have varargs.
    if f_varargs is not None:
        arguments[f_varargs] = call_args[n_f_args:]
    # set the f_kwargs key
    # --------------------
    # Again, we need a check because there can be param mismatches
    if f_kwargs is not None:
        arguments[f_kwargs] = call_kwargs
    return arguments


def _get_arguments_no_extra_positional(call_args, call_kwargs,
                                       f_args, f_defaults, f_varargs,
                                       f_kwargs):
    arguments = {}
    # if varargs exist in the signature, add them as empty
    if f_varargs is not None:
        arguments[f_varargs] = ()
    # segment the regular arguments of f into three groups
    #  * everything before end_positional is specified positionally
    #    in the call, regardless of whether it has a default value
    #  * everything between end_positional and start_defaulted must
    #    either be specified via a kwarg in the call or be missing
    #  * everything after start_defaulted has a default value and is
    #    not positional, so it either was a kwarg in the call or has the
    #    default value
    n_args = len(f_args)
    n_positional = len(call_args)
    n_no_defaults = n_args - len(f_defaults)
    for i, name in enumerate(f_args):
        # if there's a positional argument in the call, that's what we use
        if i < n_positional:
            arguments[name] = call_args[i]
        else:
            # if there's no positional arg but there is a kwarg in the call,
            # that's what we use
            if name in call_kwargs:
                arguments[name] = call_kwargs[name]
            else:
                # otherwise either we use a default from the signature of
                # f or the argument is missing.
                idx_in_defaults = i - n_no_defaults
                if idx_in_defaults >= 0:
                    arguments[name] = f_defaults[idx_in_defaults]
                else:
                    arguments[name] = '__missing_argument_{}__'.format(name)
    # set the f_kwargs key
    if f_kwargs is not None:
        arguments[f_kwargs] = {
            k: v for k, v in call_kwargs.iteritems() if k not in f_args
        }
    return arguments



from __future__ import print_function

import warnings
import time
import random

import wrapt

from .inspectcall import get_callargs
from .exceptions import lazy_error_prefix


def wraptify(decorator):
    """
    By adding a wrapping layer, make a decorator handle metadata
    using wrapt

    PARAMETERS
    ----------
    decorator : callable
        A decorator which may not use wrapt for handling
        metadata (most libraries don't) and therefore may clobber
        some metadata

    RETURNS
    -------
    wraptified : callable
        A decorator whose output is the same as that of
        `decorator`, except with an extra layer of wrapping
        that takes no actions but attaches metadata based on
        the decorator target using wrapt

    This function is very useful when working with third-party
    decorators, most of which use functools.wraps and clobber
    the argspec. This will cause problems if, for example, you
    want the argspec to still be available in ipython or if you
    want to use functions like those from `tdx.inspectcall` that
    use the argspec.

    """
    def wraptified(wrapped):
        inner_wrapper = decorator(wrapped)
        outer_wrapper = _wrapt_proxy_decorator(inner_wrapper)(wrapped)
        return outer_wrapper

    return wraptified


def _wrapt_proxy_decorator(actual_wrapper):
    """
    Given a wrapper function which is the output of a call
    to some decorator (which does not use wrapt), return
    a new decorator which calls said wrapper (rather than the
    target of the decoration) but picks up metadata from the
    target function / object

    """
    @wrapt.decorator()
    def proxied_wrapper(nominal_target, instance, args, kwargs):
        return actual_wrapper(*args, **kwargs)
    return proxied_wrapper


def retry(n_retry,
          exceptions=(Exception,),
          wait_time=1, wait_multiplier=2,
          jitter=0.1,
          no_wait_first_retry=True,
          warn=True, warn_f=None, warn_msg=None):
    """
    Decorate a function to retry on failures

    PARAMETERS
    ----------
    n_retry : int
        Number of times to retry
    exceptions : tuple, optional
        A list or tuple of exception types to catch. If not specified then
        we catch Exception (so anything). Any exception that isn't a subtype
        of something in this tuple will be raised without retry.
    no_wait_first_retry : {True, False}, optional
        Should we start waiting wait_time seconds on the first retry? By
        default the very first retry is immediate, but you can toggle this
        behavior
    wait_time : {int, float}, optional
        Seconds to wait the first time we wait between retries (whether this
        happens before the first retry or the second depends on the value
        of `no_wait_first_retry`)
    wait_multiplier : {int, float}, optional
        Multiplier for wait time, to enable exponential backoff.
    jitter : {float}, optional
        jitter on the wait. The actual wait time between retries is
        (the actual wait time) +/- (a random uniform drawn from the
        interval [ - jitter * wait_time, + jitter * wait_time ])
          The purpose of randomization is to ensure that if failures are due to
        concurrency issues (for example you have several workers whose clients
        are trying to access a lockable resource at the same time) then
        separate processes don't retry in lock-step. If you set `jitter`
        to zero, you can disable it.
    warn : {True, False}, optional
        Should we warn on failures that are being retried?
    warn_f : callable, optional
        Function to call with warning message. If `None` and `warn` is True,
        then we use `warnings.warn`
    warn_msg : string, optional
        String to use in warning message. Can use python formatting keys
        '{f}' and '{exception}' or '{exception!r}' in the message.

    RETURNS
    -------
    retry_wrapped : callable
        The original function wrapped in retry logic. The wrapping is done
        with wrapt so all the metadata should be preserved.

    """
    if not 0 <= jitter < 1:
        raise ValueError('`jitter` must be nonnegative and less than '
                         '1, got %r' % jitter)

    exceptions = tuple(exceptions)
    warn_f = warnings.warn if warn_f is None else warn_f
    warn_msg = ("Function {f} raised {exception!r}, retrying"
                if warn_msg is None else warn_msg)

    def randomized_wait(wait):
        return (1 + random.uniform(-jitter, jitter)) * wait

    @wrapt.decorator()
    def wrapper(f, instance, args, kwargs):
        wait = wait_time
        failed = 0
        while True:
            try:
                return f(*args, **kwargs)
            except exceptions as e:
                failed += 1
                if failed > n_retry:
                    raise
                if warn:
                    warn_f(warn_msg.format(f=f.__name__, exception=e))
                if failed > 1 or (not no_wait_first_retry):
                    time.sleep(randomized_wait(wait))
                    wait *= wait_multiplier

    return wrapper


def error_prefix_from_args(context_message):
    """
    Decorator for wrapping any function in a call to
    `tdxutil.exceptions.try_with_lazy_context`, with a context message that
    will be formatted based on the arguments of the call.

    PARAMETERS
    ----------
    context_message : str
        A message to prepend to any exceptions thrown by the function
        we are decorating. If an exception is thrown, we will format
        the context message with the arguments (including defaults)
        before raising.

    RETURNS
    -------
    decorator : function
        A decorator to wrap functions with error context prepending.
        If there's a problem formatting the message with the call
        arguments, the original message is prepended along with a
        message indicating a formatting error and the args / kwargs
        of the actual call.

    EXAMPLE
    -------

    >>> @error_context("In call, x={x}, y={y}, z={z}, fkwargs={fkwargs}")
    >>> def f(x, y=2, z=3, **fkwargs): raise Exception("error")

    >>> f(1, z=4, this='that')
    # ...traceback information, which is not affected...
    # Exception: In call, x=1, y=2, z=4, fkwargs={'this': 'that'}:
    #    error

    NOTE
    ----
    This decorator depends on `inspect.getargspec` returning a
    correct argspec. This will work if you're using decorators that
    make use of the `wrapt` library, but `functools.wraps` will
    clobber some of the metadata needed. And many third-party decorators
    (for example, most click decorators) use `functools`, which means
    they destroy the information needed.

    To ensure that the argument information is still available when
    this decorator runs, you must ensure that all the decorators in
    between the raw function and this use `inspectcall.wrapping.wraps`.
    Generally I recommand making this the innermost decorator.

    """

    @wrapt.decorator
    def wrapper(wrapped, instance, args, kwargs):

        def format_context_message():
            arguments = get_callargs(wrapped, *args, **kwargs)
            try:
                return context_message.format(**arguments)
            except:
                return (
                    context_message +
                    '\n...FAILED TO FORMAT args=%r kwargs=%r' % (args, kwargs)
                )

        with lazy_error_prefix(format_context_message):
            return wrapped(*args, **kwargs)

    return wrapper


def debug(debug=True, delay=0, use_debugger=None):
    """
    Decorator to add debug-on-error to a function.

    PARAMETERS
    ----------
    debug : boolean, optional
        If false, the decorator just returns f. This allows you to
        decorate functions and toggle debugging on and off using
        global configuration.
    delay : {float, int}, optional
        We delay `delay` seconds prior to entering the debugger. By
        including this in function definitions, you give users some
        time to keyboard interrupt if, by reading the exception message,
        they already know what's wrong and don't want to debug.
    use_debugger : {module or object, None}, optional
        Allows the user to set a non-default debugger, such as the
        `ipdb` module. If unspecified, we will first try to use `pudb`
        and fall back to `pdb` if we don't find it.

    RETURNS
    -------
    decorator : function
        A decorator which wraps a function with logic to drop into
        a debugger if any exceptions are raised by the target function.

    """
    @wrapt.decorator
    def wrapper(wrapped, instance, args, kwargs):
        try:
            return wrapped(*args, **kwargs)
        except:
            if not debug:
                raise
            import sys
            import traceback
            import time
            traceback.print_exc()
            print(("\nSleeping for %d seconds before debug, press C-c "
                   " to exit") % delay)
            time.sleep(delay)
            _, _, tb = sys.exc_info()
            if use_debugger is None:
                debugger = _get_debugger()
            else:
                debugger = use_debugger
            debugger.post_mortem(tb)

    return wrapper


def _get_debugger():
    """
    Get a debugger for use in the debug wrapper. Factoring this out
    of the debug wrapper makes it a bit more pluggable, especially
    for tests.
    """
    try:
        import pudb
        return pudb
    except ImportError:
        print ("Using pudb not found, using pdb for post-mortem. "
               "Try pip install pudb.")
        import pdb
        return pdb



"Utilities for handling files and data and loading"
import importlib
import os
import json
import yaml
import datetime


def load_python_object(identifier, descr):
    """
    Load a python object that's available from the current session

    Parameters
    ----------
    identifier : str
        A full python object identifier including modules, for example
        'np.linalg.lstsq'.
    descr : str
        A description of the thing you are trying to load, used to give
        better error messages if the load fails.

    Returns
    -------
    obj : object
        The loaded python object, for example if you have numpy
        installed on your system and ask for `'np.linalg.lstsq'`, we would
        return the `lstsq` function.

    Notes
    -----
    The main use case here is to allow user-input configuration, for
    example frameworks that define an interface and load user implementations
    at runtime.

    Examples
    --------
    >>> load_python_object("numpy.ndarray", "numpy ndarray function")
    """
    split_name = identifier.split('.')
    module_name = '.'.join(split_name[:-1])
    object_name = split_name[-1]
    try:
        module = importlib.import_module(module_name)
        return getattr(module, object_name)
    except ImportError:
        raise ValueError('Could not load %s module %s' % (descr, module_name))
    except AttributeError:
        raise ValueError('Could not load %r %s' % (descr, identifier))


def directory_of(filename):
    """
    Get the directory of the file specified by `filename`.

    Typically used from python modules which have resources to load,
    e.g. `THIS_DIR = this_dir(__file__)`
    """
    return os.path.dirname(os.path.abspath(filename))


def path_from(filename, *relative_path):
    """
    Given a filename (usually the __file__ of a python module) and
    a relative path to a resource from that file, return the filename
    to the relative resource.

    Since this might be used for reading or writing, we don't check that
    the resource exists.
    """
    return os.path.join(directory_of(filename), *relative_path)


def read_yaml(path):
    """
    Return the contents of a yaml file.
    """
    with open(path, 'r') as f:
        return yaml.safe_load(f)


def read_json(path):
    """
    Return the contents of a json file, assuming utf-8 encoding
    """
    with open(path, 'r') as f:
        return json.load(f, encoding='utf-8')


def write_json(content, path, compact=False):
    """
    Write `content` to a json file at `path` with utf-8 encoding
    The file must not currently exist.

    If `compact` is True, we write with no whitespace, otherwise we use json
    pretty printing, with indent of 2.

    """
    _raise_if_exists(path)
    with open(path, 'w') as f:
        if compact:
            json.dump(content, f, encoding='utf-8')
        else:
            json.dump(content, f, indent=2, encoding='utf-8')


def write_content(content, path):
    """
    Write string content to a file at `path`.
    The file must not currently exist.
    """
    _raise_if_exists(path)
    with open(path, 'w') as f:
        f.write(content)


def read_content(path):
    """
    Read the contents of a file, return as a string

    """
    with open(path, 'r') as f:
        return f.read()


def _raise_if_exists(path):
    if os.path.exists(path):
        raise ValueError(
            "tdxutil write functions may only be used for new files, file "
            "%s exists" % path
        )


# Convert datetimes to strings. From SO thread:
class DateTimeEncoder(json.JSONEncoder):
    """
    Customized json encoder that can handle datetime.date and
    datetime.datetime; they are converted to iso format

    When you decode, the result will be a string (tdx does not currently have
    auto-decoding), but it should be easily machine-convertible

    Adapted from this Stack Overflow thread (extended to dates as well as datetimes)
    http://stackoverflow.com/questions/11875770/how-to-overcome-datetime-datetime-not-json-serializable-in-python # noqa
    """

    def default(self, o):
        if isinstance(o, datetime.datetime) or isinstance(o, datetime.date):
            return o.isoformat()

        return json.JSONEncoder.default(self, o)



"""
Tools to make working with exceptions easier.
"""
import contextlib


def error_prefix(msg_prefix):
    """
    Non-lazy version of `lazy_error_context`. The `msg_prefix` argument
    should be a string.
    """
    return lazy_error_prefix(lambda: msg_prefix)


@contextlib.contextmanager
def lazy_error_prefix(msg_prefix_f):
    """
    Create an error context manager that attaches a prefix to the exception
    message and then raises (the original stack trace is preserved).

    The `error_context` argument should be a callable taking no arguments and
    returning a message which gets prepended to any errors.
    """
    try:
        yield
    except Exception as e:
        msg_prefix = msg_prefix_f()
        if len(e.args) == 0:
            e.args = (msg_prefix, )
        else:
            e.args = tuple(
                ["%s:\n%s" % (msg_prefix, e.args[0])] + [a for a in e.args[1:]]
            )
        raise



"Utilities to help with common validation tasks"

ANY = True


def check_keys(d, required, allowed=None, optional=None,
               descr="the dictionary"):
    """
    Check that the keys of dictionary `d` are as expected. Raises an error when
    there are keys mismatches, which can be customized with the `descr` input.

    PARAMETERS
    ----------
    d : dict
       A dictionary to validate
    required : iterable
       An iterable of required keys. We raise an error if `d` is missing any
       of these.
    allowed : {iterable, ANY, None}
       Either `tdxutil.validation.ANY`, an iterable of allowed keys, or None.
       If `ANY`, then we only check required keys (i.e. unknown keys do not
       cause any errors). If an iterable, then we raise an error if `d` has any
       keys not here. If `None`, then the set of allowed keys will be inferred
       from `required` and `optional`
    optional : {iterable, None}, optional
       An iterable of optional keys. You may specify this as an alternative to
       `allowed`, but you may not give non-`None` values for both of them. If
       both `allowed` and `optional` are `None`, then we will raise whenever
       `d`s keys do not match `required` exactly.
    descr : string, optional
       A description of the object being validated. Used in the error
       message.

    """
    if optional is not None and allowed is not None:
        raise ValueError("You may specify at most one of `allowed` and "
                         "`optional`")
    # figure out what keys are permitted
    if allowed == ANY:
        pass
    elif allowed is not None:
        allowed = set(allowed)
        required_but_not_allowed = [r for r in required if r not in allowed]
        if len(required_but_not_allowed) > 0:
            raise ValueError()
    elif optional is not None:
        allowed = set.union(set(optional), set(required))
    else:
        allowed = set(required)
    # check for missing required keys
    keyset = set(d.keys())
    missing = [k for k in required if k not in keyset]
    # unless `allowed` is ANY, check for unexpected keys
    unexpected = (
        [k for k in keyset if k not in allowed]
        if allowed != ANY
        else []
    )
    if len(missing) > 0 or len(unexpected) > 0:
        raise ValueError(
            "Key mismatch in %s. Missing required keys %r, "
            "found unexpected keys %r" %
            (descr, missing, unexpected)
        )


def check_all_in(collection, permissible, descr='the collection'):
    """
    Verify that all items in a collection are in some permissible set

    Parameters
    ----------
    collection : collection
        A collection of items expected to all be permissible
    permissible : {set, dict, list, tuple}
        A set or dict (dicts are allowed because it's a common use
        case to pre-validate that keys can be looked up) of permissible
        values for items in `collection`.
    descr : str
        A description of the object being validated. Used in the
        error message.

    """
    if not isinstance(permissible, set):
        permissible = set(permissible)
    unknown = {item for item in collection
               if item not in permissible}
    if len(unknown) > 0:
        raise ValueError("Unknown items %r in %s" % (unknown, descr))



# flake8: noqa
from . import data
from . import collections
from . import exceptions
from . import decorators

from .data import *
from .collections import *
from .exceptions import *
from .validation import *
from .decorators import *



import os
from setuptools import setup

PACKAGE = 'tdx'


def readme():
    with open('README.rst') as f:
        return f.read()


def get_version_str():
    ver_file = os.path.join(PACKAGE, 'version.py')
    with open(ver_file, 'r') as fid:
        version = fid.read()
    version = version.split("= ")
    version = version[1].strip()
    version = version.strip('"')
    version = version.strip("'")
    return version


# NOTE: I suspect zip_safe should actually be True, but I haven't
# had time to investigate yet.
setup(name=PACKAGE,
      version=get_version_str(),
      description=(
          'Tools for working with container types, command data operations, '
          'and concise exception handling'
      ),
      long_description=readme(),
      classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 2.7',
      ],
      keywords='',
      url='https://github.com/stroxler/tdx',
      author='Steven Troxler',
      author_email='steven.troxler@gmail.com',
      license='MIT',
      packages=[PACKAGE],
      install_requires=['pyyaml', 'wrapt>=1.10.8'],
      tests_require=['pytest'],
      include_package_data=True,
      zip_safe=False)



"""
Utilities to help with writing more declarative or concise
code using python container types.

Note: these functions are intended to enable clean code by replacing
often-repeated idioms of imperative programming. They are deliberately
simple. Neither features nor performance are important goals, and they do
little error checking.

"""
import re


class RaiseIfSpecifierNotFound(object):
    pass


def get_by_specifier(specifier, collection, default=RaiseIfSpecifierNotFound):
    """
    Given a collection (dict or list of dicts and lists -
    basically a json blob in python form), recursively find
    an entry using a specifier of the form
      "somekey.anotherkey[1].innerkey"

    For example, given the above string we would recursively
    look up
      `collection["somekey"]["anotherkey"][1]["innerkey"]

    This is likely to be particularly useful in command-line
    applications that need to grab data out of data structures
    based on bash input.

    If the specifier is invalid, we raise a ValueError. If the
    key lookup fails, we raise a ValueError if `default` is
    left at its default value of RaiseIfSpecifierNotFound, but
    otherwise we return the default.

    """
    keys = _specifier_to_keys(specifier)
    item = collection
    # TODO should we customize the message on lookup errors?
    try:
        for k in keys:
            item = item[k]
        return item
    except (IndexError, KeyError):
        if default is RaiseIfSpecifierNotFound:
            raise ValueError(
                'Failed to find key %r in specifier %r' %
                (k, specifier)
            )
        else:
            return default


def _specifier_to_keys(specifier):
    legal_specifier_rx = re.compile(r'(?:\w+|\[\d\])(?:\[\d+\]|\.\w+)*$')
    if not legal_specifier_rx.match(specifier):
        raise ValueError(
            "%r is not a legal specifier. Specifiers should "
            "look like, e.g., \"somekey[1].innerkey\"" %
            specifier
        )
    key_match_rx = re.compile(r'(\[([\d]+)\]|\w+)')
    keys = []
    for full_match, num_match in re.findall(key_match_rx, specifier):
        if num_match != '':
            keys.append(int(num_match))
        else:
            keys.append(full_match)
    return keys


def add_dicts(*args):
    """
    Combine dicts. If there are duplicate keys, raise an error.
    """
    new = {}
    for arg in args:
        for key in arg:
            if key in new:
                raise ValueError("Duplicate key: %r" % key)
            new[key] = arg[key]
    return new


def add_lists(*args):
    """
    Append lists. This is trivial but it's here for symmetry with add_dicts
    """
    out = []
    for arg in args:
        out.extend(arg)
    return out


def add_sets(*args):
    """
    Add sets. The arguments need not be sets. Returns a set of unique
    values. If the arguments include unhashable types, raises a TypeError.
    """
    out = set()
    for arg in args:
        for thing in arg:
            out.add(thing)
    return out


def flatten(*args):
    """
    Flatten a list of iterables. Returns an iterable.
    """
    for arg in args:
        for item in arg:
            yield item


def collect_by_key(pair_iter):
    """
    Collect an iterable of key, value pairs into a dict whose keys are the
    first entries in the pairs and whose values are lists of all the
    associated values, in the order they appeared.

    EXAMPLES
    >>> collect_by_key([(1, 2), (3, 4), (3, 5), (3, 5)])
    # {1: [2], 3: [4, 5, 5]}
    """
    out = {}
    for (k, v) in pair_iter:
        out[k] = out.get(k, [])
        out[k].append(v)
    return out
